@@ -1,12 +1,14 @@
-#!/usr/bin/env python3
+#!/usr/bin/env python
 # vim:fileencoding=utf-8
 # License: GPL v3 Copyright: 2017, Kovid Goyal <kovid at kovidgoyal.net>
 
 import json
 import os
 import re
+import subprocess
 import sys
-from typing import Callable, Dict, List, NamedTuple, Optional, Sequence, Tuple
+from enum import Enum
+from typing import Any, Callable, Dict, List, NamedTuple, Optional, Sequence, Tuple
 
 _plat = sys.platform.lower()
 is_linux = 'linux' in _plat
@@ -32,6 +34,19 @@ class Command(NamedTuple):
     keyfile: Optional[str] = None
 
 
+class ISA(Enum):
+    X86 = 0x03
+    AMD64 = 0x3e
+    ARM64 = 0xb7
+    Other = 0x0
+
+
+class BinaryArch(NamedTuple):
+    bits: int = 64
+    isa: ISA = ISA.AMD64
+
+
+
 class Env:
 
     cc: List[str] = []
@@ -42,6 +57,8 @@ class Env:
     ldpaths: List[str] = []
     ccver: Tuple[int, int]
     vcs_rev: str = ''
+    binary_arch: BinaryArch = BinaryArch()
+    native_optimizations: bool = False
 
     # glfw stuff
     all_headers: List[str] = []
@@ -54,22 +71,39 @@ class Env:
     def __init__(
         self, cc: List[str] = [], cppflags: List[str] = [], cflags: List[str] = [], ldflags: List[str] = [],
         library_paths: Dict[str, List[str]] = {}, ldpaths: Optional[List[str]] = None, ccver: Tuple[int, int] = (0, 0),
-        vcs_rev: str = ''
+        vcs_rev: str = '', binary_arch: BinaryArch = BinaryArch(),
+        native_optimizations: bool = False,
     ):
         self.cc, self.cppflags, self.cflags, self.ldflags, self.library_paths = cc, cppflags, cflags, ldflags, library_paths
         self.ldpaths = ldpaths or []
         self.ccver = ccver
         self.vcs_rev = vcs_rev
+        self.binary_arch = binary_arch
+        self.native_optimizations = native_optimizations
+        self._cc_version_string = ''
+
+    @property
+    def cc_version_string(self) -> str:
+        if not self._cc_version_string:
+            self._cc_version_string = subprocess.check_output(self.cc + ['--version']).decode()
+        return self._cc_version_string
+
+    @property
+    def is_gcc(self) -> bool:
+        return 'gcc' in self.cc_version_string.split(maxsplit=1)[0].lower()
 
     def copy(self) -> 'Env':
         ans = Env(self.cc, list(self.cppflags), list(self.cflags), list(self.ldflags), dict(self.library_paths), list(self.ldpaths), self.ccver)
         ans.all_headers = list(self.all_headers)
+        ans._cc_version_string = self._cc_version_string
         ans.sources = list(self.sources)
         ans.wayland_packagedir = self.wayland_packagedir
         ans.wayland_scanner = self.wayland_scanner
         ans.wayland_scanner_code = self.wayland_scanner_code
         ans.wayland_protocols = self.wayland_protocols
         ans.vcs_rev = self.vcs_rev
+        ans.binary_arch = self.binary_arch
+        ans.native_optimizations = self.native_optimizations
         return ans
 
 
@@ -83,7 +117,7 @@ def init_env(
     pkg_config: Callable[..., List[str]],
     pkg_version: Callable[[str], Tuple[int, int]],
     at_least_version: Callable[..., None],
-    test_compile: Callable[..., bool],
+    test_compile: Callable[..., Any],
     module: str = 'x11'
 ) -> Env:
     ans = env.copy()
@@ -262,6 +296,7 @@ def generate_wrappers(glfw_header: str) -> None:
     void glfwWaylandActivateWindow(GLFWwindow *handle, const char *activation_token)
     void glfwWaylandRunWithActivationToken(GLFWwindow *handle, GLFWactivationcallback cb, void *cb_data)
     bool glfwWaylandSetTitlebarColor(GLFWwindow *handle, uint32_t color, bool use_system_color)
+    void glfwWaylandRedrawCSDWindowTitle(GLFWwindow *handle)
     unsigned long long glfwDBusUserNotify(const char *app_name, const char* icon, const char *summary, const char *body, \
 const char *action_text, int32_t timeout, GLFWDBusnotificationcreatedfun callback, void *data)
     void glfwDBusSetUserNotificationHandler(GLFWDBusnotificationactivatedfun handler)
