@@ -57,7 +57,7 @@ void _glfwInputWindowFocus(_GLFWwindow* window, bool focused)
             if (window->activated_keys[i].key > 0 && window->activated_keys[i].action == GLFW_PRESS)
             {
                 const int native_key = _glfwPlatformGetNativeKeyForKey(window->activated_keys[i].key);
-                GLFWkeyevent ev = {.key = window->activated_keys[i].key, .native_key = native_key, .action = GLFW_RELEASE};
+                GLFWkeyevent ev = {.key = window->activated_keys[i].key, .native_key = native_key, .action = GLFW_RELEASE, .fake_event_on_focus_change = true};
                 _glfwInputKeyboard(window, &ev);
             }
         }
@@ -334,6 +334,9 @@ void glfwDefaultWindowHints(void)
     _glfw.hints.window.ns.retina = true;
     // use the default colorspace assigned by the system
     _glfw.hints.window.ns.color_space = 0;
+    // no blur
+    _glfw.hints.window.ns.blur_radius = 0;
+    _glfw.hints.window.x11.enable_blur = 0;
 }
 
 GLFWAPI void glfwWindowHint(int hint, int value)
@@ -417,6 +420,12 @@ GLFWAPI void glfwWindowHint(int hint, int value)
         case GLFW_COCOA_COLOR_SPACE:
             _glfw.hints.window.ns.color_space = value;
             return;
+        case GLFW_COCOA_BLUR_RADIUS:
+            _glfw.hints.window.ns.blur_radius = value;
+            return;
+        case GLFW_X11_BLUR:
+            _glfw.hints.window.x11.enable_blur = value;
+            return;
         case GLFW_COCOA_GRAPHICS_SWITCHING:
             _glfw.hints.context.nsgl.offline = value ? true : false;
             return;
