@@ -11,11 +11,9 @@
 #include "threading.h"
 #include "screen.h"
 #include "fonts.h"
-#include "charsets.h"
 #include "monotonic.h"
 #include <termios.h>
 #include <unistd.h>
-#include <float.h>
 #include <fcntl.h>
 #include <sys/ioctl.h>
 #include <sys/wait.h>
@@ -41,12 +39,11 @@ extern PyTypeObject Screen_Type;
 #endif
 #define USE_RENDER_FRAMES (global_state.has_render_frames && OPT(sync_to_monitor))
 
-static void (*parse_func)(Screen*, PyObject*, monotonic_t);
-
 typedef struct {
     char *data;
     size_t sz;
     id_type peer_id;
+    bool is_remote_control_peer;
 } Message;
 
 typedef struct {
@@ -61,6 +58,7 @@ typedef struct {
     Message *messages;
     size_t messages_capacity, messages_count;
     LoopData io_loop_data;
+    void (*parse_func)(void*, ParseData*, bool);
 } ChildMonitor;
 
 
@@ -155,7 +153,7 @@ mask_kitty_signals_process_wide(PyObject *self UNUSED, PyObject *a UNUSED) {
 }
 
 static PyObject *
-new(PyTypeObject *type, PyObject *args, PyObject UNUSED *kwds) {
+new_childmonitor_object(PyTypeObject *type, PyObject *args, PyObject UNUSED *kwds) {
     ChildMonitor *self;
     PyObject *dump_callback, *death_notify;
     int talk_fd = -1, listen_fd = -1;
@@ -179,8 +177,8 @@ new(PyTypeObject *type, PyObject *args, PyObject UNUSED *kwds) {
     self->death_notify = death_notify; Py_INCREF(death_notify);
     if (dump_callback != Py_None) {
         self->dump_callback = dump_callback; Py_INCREF(dump_callback);
-        parse_func = parse_worker_dump;
-    } else parse_func = parse_worker;
+        self->parse_func = parse_worker_dump;
+    } else self->parse_func = parse_worker;
     self->count = 0;
     children_fds[0].fd = self->io_loop_data.wakeup_read_fd; children_fds[1].fd = self->io_loop_data.signal_read_fd;
     children_fds[0].events = POLLIN; children_fds[1].events = POLLIN; children_fds[2].events = POLLIN;
@@ -232,8 +230,53 @@ static void* io_loop(void *data);
 static void* talk_loop(void *data);
 static void send_response_to_peer(id_type peer_id, const char *msg, size_t msg_sz);
 static void wakeup_talk_loop(bool);
+static bool add_peer_to_injection_queue(int peer_fd, int pipe_fd);
 static bool talk_thread_started = false;
 
+static bool
+simple_read_from_pipe(int fd, void *data, size_t sz) {
+    // read a small amount of data to a pipe handling only EINTR
+    while (true) {
+        ssize_t ret = read(fd, data, sz);
+        if (ret == -1 && errno == EINTR) continue;
+        return ret == (ssize_t)sz;
+    }
+}
+
+
+static PyObject*
+inject_peer(PyObject *s, PyObject *a) {
+#define inject_peer_doc "inject_peer(fd) -> Start communication with a peer over the specified file descriptor"
+    ChildMonitor *self = (ChildMonitor*)s;
+    if (!PyLong_Check(a)) { PyErr_SetString(PyExc_TypeError, "peer fd must be an int"); return NULL; }
+    long fd = PyLong_AsLong(a);
+    if (fd < 0) { PyErr_Format(PyExc_ValueError, "Invalid peer fd: %ld", fd); return NULL; }
+    if (!talk_thread_started) {
+        int ret;
+        if ((ret = pthread_create(&self->talk_thread, NULL, talk_loop, self)) != 0) {
+            return PyErr_Format(PyExc_OSError, "Failed to start talk thread with error: %s", strerror(ret));
+        }
+        talk_thread_started = true;
+    }
+    int fds[2] = {0};
+    if (!self_pipe(fds, false)) {
+        safe_close(fd, __FILE__, __LINE__);
+        return PyErr_SetFromErrno(PyExc_OSError);
+    }
+    if (!add_peer_to_injection_queue(fd, fds[1])) {
+        safe_close(fd, __FILE__, __LINE__);
+        safe_close(fds[0], __FILE__, __LINE__); safe_close(fds[1], __FILE__, __LINE__);
+        PyErr_SetString(PyExc_RuntimeError, "Too many peers waiting to be injected");
+        return NULL;
+    }
+    wakeup_talk_loop(false);
+    id_type peer_id = 0;
+    bool ok = simple_read_from_pipe(fds[0], &peer_id, sizeof(peer_id));
+    safe_close(fds[0], __FILE__, __LINE__);
+    if (!ok) { PyErr_SetString(PyExc_RuntimeError, "Failed to read peer id from self pipe"); return NULL; }
+    return PyLong_FromUnsignedLongLong(peer_id);
+}
+
 static PyObject *
 start(PyObject *s, PyObject *a UNUSED) {
 #define start_doc "start() -> Start the I/O thread"
@@ -393,24 +436,15 @@ shutdown_monitor(ChildMonitor *self, PyObject *a UNUSED) {
 
 static bool
 do_parse(ChildMonitor *self, Screen *screen, monotonic_t now, bool flush) {
-    bool input_read = false;
-    screen_mutex(lock, read);
-    if (screen->read_buf_sz || screen->pending_mode.used) {
-        monotonic_t time_since_new_input = now - screen->new_input_at;
-        if (flush || time_since_new_input >= OPT(input_delay)) {
-            bool read_buf_full = screen->read_buf_sz >= READ_BUF_SZ;
-            input_read = true;
-            parse_func(screen, self->dump_callback, now);
-            if (read_buf_full) wakeup_io_loop(self, false);  // Ensure the read fd has POLLIN set
-            screen->new_input_at = 0;
-            if (screen->pending_mode.activated_at) {
-                monotonic_t time_since_pending = MAX(0, now - screen->pending_mode.activated_at);
-                set_maximum_wait(screen->pending_mode.wait_time - time_since_pending);
-            }
-        } else set_maximum_wait(OPT(input_delay) - time_since_new_input);
-    }
-    screen_mutex(unlock, read);
-    return input_read;
+    ParseData pd = {.dump_callback = self->dump_callback, .now = now};
+    self->parse_func(screen, &pd, flush);
+    if (pd.input_read) {
+        if (pd.write_space_created) wakeup_io_loop(self, false);
+        if (screen->paused_rendering.expires_at) {
+            set_maximum_wait(MAX(0, screen->paused_rendering.expires_at - now));
+        } else set_maximum_wait(OPT(input_delay) - pd.time_since_new_input);
+    } else if (pd.has_pending_input) set_maximum_wait(OPT(input_delay) - pd.time_since_new_input);
+    return pd.input_read;
 }
 
 static bool
@@ -467,7 +501,7 @@ parse_input(ChildMonitor *self) {
             Message *msg = msgs + i;
             PyObject *resp = NULL;
             if (msg->data) {
-                resp = PyObject_CallMethod(global_state.boss, "peer_message_received", "y#K", msg->data, (int)msg->sz, msg->peer_id);
+                resp = PyObject_CallMethod(global_state.boss, "peer_message_received", "y#KO", msg->data, (int)msg->sz, msg->peer_id, msg->is_remote_control_peer ? Py_True : Py_False);
                 free(msg->data);
                 if (!resp) PyErr_Print();
             }
@@ -617,12 +651,14 @@ pyset_iutf8(ChildMonitor *self, PyObject *args) {
 
 static bool
 cursor_needs_render(Window *w) {
-    return w->cursor_visible_at_last_render != w->render_data.screen->cursor_render_info.is_visible || w->last_cursor_x != w->render_data.screen->cursor_render_info.x || w->last_cursor_y != w->render_data.screen->cursor_render_info.y || w->last_cursor_shape != w->render_data.screen->cursor_render_info.shape;
+#define cri w->render_data.screen->cursor_render_info
+    return w->cursor_visible_at_last_render != cri.is_visible || w->render_data.screen->last_rendered.cursor_x != cri.x || w->render_data.screen->last_rendered.cursor_y != cri.y || w->last_cursor_shape != cri.shape;
+#undef cri
 }
 
 static bool
 collect_cursor_info(CursorRenderInfo *ans, Window *w, monotonic_t now, OSWindow *os_window) {
-    ScreenRenderData *rd = &w->render_data;
+    WindowRenderData *rd = &w->render_data;
     const Cursor *cursor;
     if (screen_is_overlay_active(rd->screen)) {
         // Do not force the cursor to be visible here for the sake of some programs that prefer it hidden
@@ -630,7 +666,7 @@ collect_cursor_info(CursorRenderInfo *ans, Window *w, monotonic_t now, OSWindow
         ans->x = rd->screen->overlay_line.cursor_x;
         ans->y = rd->screen->overlay_line.ynum;
     } else {
-        cursor = rd->screen->cursor;
+        cursor = rd->screen->paused_rendering.expires_at ? &rd->screen->paused_rendering.cursor : rd->screen->cursor;
         ans->x = cursor->x; ans->y = cursor->y;
     }
     ans->is_visible = false;
@@ -690,6 +726,7 @@ prepare_to_render_os_window(OSWindow *os_window, monotonic_t now, unsigned int *
         Window *w = tab->windows + i;
 #define WD w->render_data
         if (w->visible && WD.screen) {
+            screen_check_pause_rendering(WD.screen, now);
             *num_visible_windows += 1;
             color_type window_bg = colorprofile_to_color(WD.screen->color_profile, WD.screen->color_profile->overridden.default_bg, WD.screen->color_profile->configured.default_bg).rgb;
             if (*num_visible_windows == 1) first_window_bg = window_bg;
@@ -711,7 +748,7 @@ prepare_to_render_os_window(OSWindow *os_window, monotonic_t now, unsigned int *
                 set_os_window_title_from_window(w, os_window);
                 *active_window_bg = window_bg;
             } else {
-                if (WD.screen->render_unfocused_cursor) {
+                if (WD.screen->cursor_render_info.render_even_when_unfocused) {
                     if (collect_cursor_info(&WD.screen->cursor_render_info, w, now, os_window)) needs_render = true;
                     WD.screen->cursor_render_info.is_focused = false;
                 } else {
@@ -764,7 +801,7 @@ render_prepared_os_window(OSWindow *os_window, unsigned int active_window_id, co
             if (WD.screen->start_visual_bell_at != 0) {
                 set_maximum_wait(OPT(repaint_delay));
             }
-            w->cursor_visible_at_last_render = WD.screen->cursor_render_info.is_visible; w->last_cursor_x = WD.screen->cursor_render_info.x; w->last_cursor_y = WD.screen->cursor_render_info.y; w->last_cursor_shape = WD.screen->cursor_render_info.shape;
+            w->cursor_visible_at_last_render = WD.screen->cursor_render_info.is_visible; w->last_cursor_shape = WD.screen->cursor_render_info.shape;
         }
     }
     if (os_window->live_resize.in_progress) draw_resizing_text(os_window);
@@ -1043,6 +1080,10 @@ close_os_window(ChildMonitor *self, OSWindow *os_window) {
     if (os_window->before_fullscreen.is_set && is_os_window_fullscreen(os_window)) {
         w = os_window->before_fullscreen.w; h = os_window->before_fullscreen.h;
     }
+    // On GNOME Wayland w, h are the content area size, we need to add the frame size back
+    int content_area_width, content_area_height;
+    adjust_window_size_for_csd(os_window, w, h, &content_area_width, &content_area_height);
+    w += w - content_area_width; h += h - content_area_height;
     destroy_os_window(os_window);
     call_boss(on_os_window_closed, "Kii", os_window->id, w, h);
     for (size_t t=0; t < os_window->num_tabs; t++) {
@@ -1148,6 +1189,7 @@ process_cocoa_pending_actions(void) {
     if (cocoa_pending_actions_data.wd) {
         if (cocoa_pending_actions[NEW_OS_WINDOW_WITH_WD]) { call_boss(new_os_window_with_wd, "sO", cocoa_pending_actions_data.wd, Py_True); }
         if (cocoa_pending_actions[NEW_TAB_WITH_WD]) { call_boss(new_tab_with_wd, "sO", cocoa_pending_actions_data.wd, Py_True); }
+        if (cocoa_pending_actions[USER_MENU_ACTION]) { call_boss(user_menu_action, "s", cocoa_pending_actions_data.wd); }
         free(cocoa_pending_actions_data.wd);
         cocoa_pending_actions_data.wd = NULL;
     }
@@ -1293,34 +1335,23 @@ remove_children(ChildMonitor *self) {
 static bool
 read_bytes(int fd, Screen *screen) {
     ssize_t len;
-    size_t available_buffer_space, orig_sz;
+    size_t available_buffer_space;
 
-    screen_mutex(lock, read);
-    orig_sz = screen->read_buf_sz;
-    if (orig_sz >= READ_BUF_SZ) { screen_mutex(unlock, read); return true; }  // screen read buffer is full
-    available_buffer_space = READ_BUF_SZ - orig_sz;
-    screen_mutex(unlock, read);
+    uint8_t *buf = vt_parser_create_write_buffer(screen->vt_parser, &available_buffer_space);
+    if (!available_buffer_space) return true;
 
     while(true) {
-        len = read(fd, screen->read_buf + orig_sz, available_buffer_space);
+        len = read(fd, buf, available_buffer_space);
         if (len < 0) {
             if (errno == EINTR || errno == EAGAIN) continue;
             if (errno != EIO) perror("Call to read() from child fd failed");
+            vt_parser_commit_write(screen->vt_parser, 0);
             return false;
         }
         break;
     }
-    if (UNLIKELY(len == 0)) return false;
-
-    screen_mutex(lock, read);
-    if (screen->new_input_at == 0) screen->new_input_at = monotonic();
-    if (orig_sz != screen->read_buf_sz) {
-        // The other thread consumed some of the screen read buffer
-        memmove(screen->read_buf + screen->read_buf_sz, screen->read_buf + orig_sz, len);
-    }
-    screen->read_buf_sz += len;
-    screen_mutex(unlock, read);
-    return true;
+    vt_parser_commit_write(screen->vt_parser, len);
+    return len != 0;
 }
 
 
@@ -1466,9 +1497,10 @@ io_loop(void *data) {
         for (i = 0; i < self->count; i++) {
             screen = children[i].screen;
             /* printf("i:%lu id:%lu fd: %d read_buf_sz: %lu write_buf_used: %lu\n", i, children[i].id, children[i].fd, screen->read_buf_sz, screen->write_buf_used); */
-            screen_mutex(lock, read); screen_mutex(lock, write);
-            children_fds[EXTRA_FDS + i].events = (screen->read_buf_sz < READ_BUF_SZ ? POLLIN : 0) | (screen->write_buf_used ? POLLOUT  : 0);
-            screen_mutex(unlock, read); screen_mutex(unlock, write);
+            children_fds[EXTRA_FDS + i].events = vt_parser_has_space_for_input(screen->vt_parser) ? POLLIN : 0;
+            screen_mutex(lock, write);
+            children_fds[EXTRA_FDS + i].events |= (screen->write_buf_used ? POLLOUT  : 0);
+            screen_mutex(unlock, write);
         }
         if (has_pending_wakeups) {
             now = monotonic();
@@ -1562,6 +1594,7 @@ typedef struct {
         size_t capacity, used;
         bool failed;
     } write;
+    bool is_remote_control_peer;
 } Peer;
 static id_type peer_id_counter = 0;
 
@@ -1576,24 +1609,33 @@ typedef struct pollfd PollFD;
 #define PEER_LIMIT 256
 #define nuke_socket(s) { shutdown(s, SHUT_RDWR); safe_close(s, __FILE__, __LINE__); }
 
-static bool
-accept_peer(int listen_fd, bool shutting_down) {
-    int peer = accept(listen_fd, NULL, NULL);
-    if (UNLIKELY(peer == -1)) {
-        if (errno == EINTR) return true;
-        if (!shutting_down) perror("accept() on talk socket failed!");
-        return false;
-    }
+static id_type
+add_peer(int peer, bool is_remote_control_peer) {
+    id_type ans = 0;
     if (talk_data.num_peers < PEER_LIMIT) {
         ensure_space_for(&talk_data, peers, Peer, talk_data.num_peers + 8, peers_capacity, 8, false);
         Peer *p = talk_data.peers + talk_data.num_peers++;
         memset(p, 0, sizeof(Peer));
         p->fd = peer; p->id = ++peer_id_counter;
         if (!p->id) p->id = ++peer_id_counter;
+        ans = p->id;
+        p->is_remote_control_peer = is_remote_control_peer;
     } else {
         log_error("Too many peers want to talk, ignoring one.");
         nuke_socket(peer);
     }
+    return ans;
+}
+
+static bool
+accept_peer(int listen_fd, bool shutting_down, bool is_remote_control_peer) {
+    int peer = accept(listen_fd, NULL, NULL);
+    if (UNLIKELY(peer == -1)) {
+        if (errno == EINTR) return true;
+        if (!shutting_down) perror("accept() on talk socket failed!");
+        return false;
+    }
+    add_peer(peer, is_remote_control_peer);
     return true;
 }
 
@@ -1620,11 +1662,23 @@ queue_peer_message(ChildMonitor *self, Peer *peer) {
         }
     }
     m->peer_id = peer->id;
+    m->is_remote_control_peer = peer->is_remote_control_peer;
     peer->num_of_unresponded_messages_sent_to_main_thread++;
     talk_mutex(unlock);
     wakeup_main_loop();
 }
 
+static void
+notify_on_peer_removal(ChildMonitor *self, const Peer *p) {
+    ensure_space_for(self, messages, Message, self->messages_count + 16, messages_capacity, 16, true);
+    Message *m = self->messages + self->messages_count++;
+    memset(m, 0, sizeof(Message));
+    m->data = strdup("peer_death");
+    if (m->data) m->sz = strlen("peer_death");
+    m->peer_id = p->id;
+    m->is_remote_control_peer = p->id;
+}
+
 static bool
 has_complete_peer_command(Peer *peer) {
     peer->read.command_end = 0;
@@ -1701,17 +1755,51 @@ wakeup_talk_loop(bool in_signal_handler) {
 }
 
 
-static void
-prune_peers(void) {
+static bool
+prune_peers(ChildMonitor *self) {
+    bool pruned = false;
     for (size_t idx = talk_data.num_peers; idx-- > 0;) {
         Peer *p = talk_data.peers + idx;
         if (p->read.finished && !p->num_of_unresponded_messages_sent_to_main_thread && !p->write.used) {
+            notify_on_peer_removal(self, p);
             free_peer(p);
             remove_i_from_array(talk_data.peers, idx, talk_data.num_peers);
+            pruned = true;
         }
     }
+    return pruned;
 }
 
+static struct {
+    size_t num;
+    struct { int peer_fd, pipe_fd; } fds[16];
+} peers_to_inject = {0};
+
+static bool
+add_peer_to_injection_queue(int peer_fd, int pipe_fd) {
+    bool added = false;
+    talk_mutex(lock);
+    if (peers_to_inject.num < arraysz(peers_to_inject.fds)) {
+        peers_to_inject.fds[peers_to_inject.num].peer_fd = peer_fd;
+        peers_to_inject.fds[peers_to_inject.num].pipe_fd = pipe_fd;
+        peers_to_inject.num++;
+        added = true;
+    }
+    talk_mutex(unlock);
+    return added;
+}
+
+static void
+simple_write_to_pipe(int fd, void *data, size_t sz) {
+    // write a small amount of data to a pipe handling only EINTR
+    while (true) {
+        ssize_t ret = write(fd, data, sz);
+        if (ret == -1 && errno == EINTR) continue;
+        break;
+    }
+}
+
+
 static void*
 talk_loop(void *data) {
     // The talk thread loop
@@ -1730,9 +1818,18 @@ talk_loop(void *data) {
 
     while (LIKELY(!self->shutting_down)) {
         num_peer_fds = 0;
+        bool need_to_wakup_main_loop = false;
+        talk_mutex(lock);
+        if (peers_to_inject.num) {
+            for (size_t i = 0; i < peers_to_inject.num; i++) {
+                id_type added_peer_id = add_peer(peers_to_inject.fds[i].peer_fd, true);
+                simple_write_to_pipe(peers_to_inject.fds[i].pipe_fd, &added_peer_id, sizeof(id_type));
+                safe_close(peers_to_inject.fds[i].pipe_fd, __FILE__, __LINE__);
+            }
+            peers_to_inject.num = 0;
+        }
         if (talk_data.num_peers > 0) {
-            talk_mutex(lock);
-            prune_peers();
+            if (prune_peers(self)) need_to_wakup_main_loop = true;
             for (size_t i = 0; i < talk_data.num_peers; i++) {
                 Peer *p = talk_data.peers + i;
                 if (!p->read.finished || p->write.used) {
@@ -1745,14 +1842,15 @@ talk_loop(void *data) {
                     fds[p->fd_array_idx].events = flags;
                 } else p->fd_array_idx = 0;
             }
-            talk_mutex(unlock);
         }
+        talk_mutex(unlock);
+        if (need_to_wakup_main_loop) wakeup_main_loop();
         for (size_t i = 0; i < num_listen_fds; i++) fds[i].revents = 0;
         int ret = poll(fds, num_listen_fds + num_peer_fds, -1);
         if (ret > 0) {
             for (size_t i = 0; i < num_listen_fds - 1; i++) {
                 if (fds[i].revents & POLLIN) {
-                    if (!accept_peer(fds[i].fd, self->shutting_down)) goto end;
+                    if (!accept_peer(fds[i].fd, self->shutting_down, fds[i].fd == self->listen_fd)) goto end;
                 }
             }
             if (fds[num_listen_fds - 1].revents & POLLIN) {
@@ -1813,6 +1911,7 @@ send_response_to_peer(id_type peer_id, const char *msg, size_t msg_sz) {
 // Boilerplate {{{
 static PyMethodDef methods[] = {
     METHOD(add_child, METH_VARARGS)
+    METHOD(inject_peer, METH_O)
     METHOD(needs_write, METH_VARARGS)
     METHOD(start, METH_NOARGS)
     METHOD(wakeup, METH_NOARGS)
@@ -1834,7 +1933,7 @@ PyTypeObject ChildMonitor_Type = {
     .tp_flags = Py_TPFLAGS_DEFAULT,
     .tp_doc = "ChildMonitor",
     .tp_methods = methods,
-    .tp_new = new,
+    .tp_new = new_childmonitor_object,
 };
 
 
