@@ -13,6 +13,7 @@
 #endif
 
 #include "data-types.h"
+#include "charsets.h"
 #include "base64.h"
 #include <sys/socket.h>
 #include <sys/types.h>
@@ -25,7 +26,6 @@
 #include "modes.h"
 #include <stddef.h>
 #include <termios.h>
-#include <signal.h>
 #include <fcntl.h>
 #include <stdio.h>
 #include <locale.h>
@@ -96,7 +96,11 @@ pybase64_decode(PyObject UNUSED *self, PyObject *args) {
     size_t sz = required_buffer_size_for_base64_decode(view.len);
     PyObject *ans = PyBytes_FromStringAndSize(NULL, sz);
     if (!ans) return NULL;
-    base64_decode8(view.buf, view.len, (unsigned char*)PyBytes_AS_STRING(ans), &sz);
+    if (!base64_decode8(view.buf, view.len, (unsigned char*)PyBytes_AS_STRING(ans), &sz)) {
+        Py_DECREF(ans);
+        PyErr_SetString(PyExc_ValueError, "Invalid base64 input data");
+        return NULL;
+    }
     if (_PyBytes_Resize(&ans, sz) != 0) return NULL;
     return ans;
 }
@@ -321,7 +325,87 @@ expand_ansi_c_escapes(PyObject *self UNUSED, PyObject *src) {
     return ans;
 }
 
+START_ALLOW_CASE_RANGE
+static PyObject*
+c0_replace_bytes(const char *input_data, Py_ssize_t input_sz) {
+    RAII_PyObject(ans, PyBytes_FromStringAndSize(NULL, input_sz * 3));
+    if (!ans) return NULL;
+    char *output = PyBytes_AS_STRING(ans);
+    char buf[4];
+    Py_ssize_t j = 0;
+    for (Py_ssize_t i = 0; i < input_sz; i++) {
+        const char x = input_data[i];
+        switch (x) {
+            case C0_EXCEPT_NL_SPACE_TAB: {
+                const uint32_t ch = 0x2400 + x;
+                const unsigned sz = encode_utf8(ch, buf);
+                for (unsigned c = 0; c < sz; c++, j++) output[j] = buf[c];
+            } break;
+            default:
+                output[j++] = x; break;
+        }
+    }
+    if (_PyBytes_Resize(&ans, j) != 0) return NULL;
+    Py_INCREF(ans);
+    return ans;
+}
+
+static PyObject*
+c0_replace_unicode(PyObject *input) {
+    RAII_PyObject(ans, PyUnicode_New(PyUnicode_GET_LENGTH(input), 1114111));
+    if (!ans) return NULL;
+    void *input_data = PyUnicode_DATA(input);
+    int input_kind = PyUnicode_KIND(input);
+    void *output_data = PyUnicode_DATA(ans);
+    int output_kind = PyUnicode_KIND(ans);
+    Py_UCS4 maxchar = 0;
+    bool changed = false;
+    for (Py_ssize_t i = 0; i < PyUnicode_GET_LENGTH(input); i++) {
+        Py_UCS4 ch = PyUnicode_READ(input_kind, input_data, i);
+        switch(ch) { case C0_EXCEPT_NL_SPACE_TAB: ch += 0x2400; changed = true; }
+        if (ch > maxchar) maxchar = ch;
+        PyUnicode_WRITE(output_kind, output_data, i, ch);
+    }
+    if (!changed) { Py_INCREF(input); return input; }
+    if (maxchar > 65535) { Py_INCREF(ans); return ans; }
+    RAII_PyObject(ans2, PyUnicode_New(PyUnicode_GET_LENGTH(ans), maxchar));
+    if (!ans2) return NULL;
+    if (PyUnicode_CopyCharacters(ans2, 0, ans, 0, PyUnicode_GET_LENGTH(ans)) == -1) return NULL;
+    Py_INCREF(ans2); return ans2;
+}
+END_ALLOW_CASE_RANGE
+
+static PyObject*
+replace_c0_codes_except_nl_space_tab(PyObject *self UNUSED, PyObject *obj) {
+    if (PyUnicode_Check(obj)) {
+        return c0_replace_unicode(obj);
+    } else if (PyBytes_Check(obj)) {
+        return c0_replace_bytes(PyBytes_AS_STRING(obj), PyBytes_GET_SIZE(obj));
+    } else if (PyMemoryView_Check(obj)) {
+        Py_buffer *buf = PyMemoryView_GET_BUFFER(obj);
+        return c0_replace_bytes(buf->buf, buf->len);
+    } else if (PyByteArray_Check(obj)) {
+        return c0_replace_bytes(PyByteArray_AS_STRING(obj), PyByteArray_GET_SIZE(obj));
+    } else {
+        PyErr_SetString(PyExc_TypeError, "Input must be bytes, memoryview, bytearray or unicode");
+        return NULL;
+    }
+}
+
+
+static PyObject*
+find_in_memoryview(PyObject *self UNUSED, PyObject *args) {
+    unsigned char q;
+    RAII_PY_BUFFER(view);
+    if (!PyArg_ParseTuple(args, "y*b", &view, &q)) return NULL;
+    const char *buf = view.buf, *p = memchr(buf, q, view.len);
+    Py_ssize_t ans = -1;
+    if (p) ans = p - buf;
+    return PyLong_FromSsize_t(ans);
+}
+
 static PyMethodDef module_methods[] = {
+    METHODB(replace_c0_codes_except_nl_space_tab, METH_O),
     {"wcwidth", (PyCFunction)wcwidth_wrap, METH_O, ""},
     {"expand_ansi_c_escapes", (PyCFunction)expand_ansi_c_escapes, METH_O, ""},
     {"get_docs_ref_map", (PyCFunction)get_docs_ref_map, METH_NOARGS, ""},
@@ -336,13 +420,12 @@ static PyMethodDef module_methods[] = {
     {"base64_encode", (PyCFunction)pybase64_encode, METH_VARARGS, ""},
     {"base64_decode", (PyCFunction)pybase64_decode, METH_VARARGS, ""},
     {"thread_write", (PyCFunction)cm_thread_write, METH_VARARGS, ""},
-    {"parse_bytes", (PyCFunction)parse_bytes, METH_VARARGS, ""},
-    {"parse_bytes_dump", (PyCFunction)parse_bytes_dump, METH_VARARGS, ""},
     {"redirect_std_streams", (PyCFunction)redirect_std_streams, METH_VARARGS, ""},
     {"locale_is_valid", (PyCFunction)locale_is_valid, METH_VARARGS, ""},
     {"shm_open", (PyCFunction)py_shm_open, METH_VARARGS, ""},
     {"shm_unlink", (PyCFunction)py_shm_unlink, METH_VARARGS, ""},
     {"wrapped_kitten_names", (PyCFunction)wrapped_kittens, METH_NOARGS, ""},
+    {"find_in_memoryview", (PyCFunction)find_in_memoryview, METH_VARARGS, ""},
 #ifdef __APPLE__
     METHODB(user_cache_dir, METH_NOARGS),
     METHODB(process_group_map, METH_NOARGS),
@@ -367,6 +450,8 @@ static struct PyModuleDef module = {
 extern int init_LineBuf(PyObject *);
 extern int init_HistoryBuf(PyObject *);
 extern int init_Cursor(PyObject *);
+extern int init_Shlex(PyObject *);
+extern int init_Parser(PyObject *);
 extern int init_DiskCache(PyObject *);
 extern bool init_child_monitor(PyObject *);
 extern int init_Line(PyObject *);
@@ -430,6 +515,8 @@ PyInit_fast_data_types(void) {
     if (!init_HistoryBuf(m)) return NULL;
     if (!init_Line(m)) return NULL;
     if (!init_Cursor(m)) return NULL;
+    if (!init_Shlex(m)) return NULL;
+    if (!init_Parser(m)) return NULL;
     if (!init_DiskCache(m)) return NULL;
     if (!init_child_monitor(m)) return NULL;
     if (!init_ColorProfile(m)) return NULL;
@@ -478,11 +565,12 @@ PyInit_fast_data_types(void) {
     PyModule_AddIntMacro(m, DECCOLM);
     PyModule_AddIntMacro(m, DECOM);
     PyModule_AddIntMacro(m, IRM);
-    PyModule_AddIntMacro(m, CSI);
-    PyModule_AddIntMacro(m, DCS);
-    PyModule_AddIntMacro(m, APC);
-    PyModule_AddIntMacro(m, OSC);
     PyModule_AddIntMacro(m, FILE_TRANSFER_CODE);
+    PyModule_AddIntMacro(m, ESC_CSI);
+    PyModule_AddIntMacro(m, ESC_OSC);
+    PyModule_AddIntMacro(m, ESC_APC);
+    PyModule_AddIntMacro(m, ESC_DCS);
+    PyModule_AddIntMacro(m, ESC_PM);
 #ifdef __APPLE__
     // Apple says its SHM_NAME_MAX but SHM_NAME_MAX is not actually declared in typical CrApple style.
     // This value is based on experimentation and from qsharedmemory.cpp in Qt
