@@ -42,7 +42,9 @@
 #define arraysz(x) (sizeof(x)/sizeof(x[0]))
 #define zero_at_i(array, idx) memset((array) + (idx), 0, sizeof((array)[0]))
 #define zero_at_ptr(p) memset((p), 0, sizeof((p)[0]))
+#define literal_strlen(x) (sizeof(x)-1)
 #define zero_at_ptr_count(p, count) memset((p), 0, (count) * sizeof((p)[0]))
+#define C0_EXCEPT_NL_SPACE_TAB 0x0 ... 0x8: case 0xb ... 0x1f: case 0x7f
 void log_error(const char *fmt, ...) __attribute__ ((format (printf, 1, 2)));
 #define fatal(...) { log_error(__VA_ARGS__); exit(EXIT_FAILURE); }
 static inline void cleanup_free(void *p) { free(*(void**)p); }
@@ -68,7 +70,41 @@ typedef enum { DISABLE_LIGATURES_NEVER, DISABLE_LIGATURES_CURSOR, DISABLE_LIGATU
 #define ERROR_PREFIX "[PARSE ERROR]"
 typedef enum MouseTrackingModes { NO_TRACKING, BUTTON_MODE, MOTION_MODE, ANY_MODE } MouseTrackingMode;
 typedef enum MouseTrackingProtocols { NORMAL_PROTOCOL, UTF8_PROTOCOL, SGR_PROTOCOL, URXVT_PROTOCOL, SGR_PIXEL_PROTOCOL} MouseTrackingProtocol;
-typedef enum MouseShapes { BEAM, HAND, ARROW } MouseShape;
+typedef enum MouseShapes {
+    INVALID_POINTER,
+    /* start mouse shapes (auto generated by gen-key-constants.py do not edit) */
+    DEFAULT_POINTER,
+    TEXT_POINTER,
+    POINTER_POINTER,
+    HELP_POINTER,
+    WAIT_POINTER,
+    PROGRESS_POINTER,
+    CROSSHAIR_POINTER,
+    CELL_POINTER,
+    VERTICAL_TEXT_POINTER,
+    MOVE_POINTER,
+    E_RESIZE_POINTER,
+    NE_RESIZE_POINTER,
+    NW_RESIZE_POINTER,
+    N_RESIZE_POINTER,
+    SE_RESIZE_POINTER,
+    SW_RESIZE_POINTER,
+    S_RESIZE_POINTER,
+    W_RESIZE_POINTER,
+    EW_RESIZE_POINTER,
+    NS_RESIZE_POINTER,
+    NESW_RESIZE_POINTER,
+    NWSE_RESIZE_POINTER,
+    ZOOM_IN_POINTER,
+    ZOOM_OUT_POINTER,
+    ALIAS_POINTER,
+    COPY_POINTER,
+    NOT_ALLOWED_POINTER,
+    NO_DROP_POINTER,
+    GRAB_POINTER,
+    GRABBING_POINTER,
+/* end mouse shapes */
+} MouseShape;
 typedef enum { NONE, MENUBAR, WINDOW, ALL } WindowTitleIn;
 typedef enum { TILING, SCALED, MIRRORED, CLAMPED, CENTER_CLAMPED, CENTER_SCALED } BackgroundImageLayout;
 typedef struct ImageAnchorPosition {
@@ -129,16 +165,17 @@ typedef struct ImageAnchorPosition {
     }
 
 #ifdef __clang__
-#define IGNORE_PEDANTIC_WARNINGS _Pragma("clang diagnostic push") _Pragma("clang diagnostic ignored \"-Wpedantic\"")
-#define END_IGNORE_PEDANTIC_WARNINGS _Pragma("clang diagnostic pop")
-#define ALLOW_UNUSED_RESULT _Pragma("clang diagnostic push") _Pragma("clang diagnostic ignored \"-Wunused-result\"")
-#define END_ALLOW_UNUSED_RESULT _Pragma("clang diagnostic pop")
+#define START_IGNORE_DIAGNOSTIC(diag) _Pragma(xstr(clang diagnostic push))  _Pragma(xstr(clang diagnostic ignored diag))
+#define END_IGNORE_DIAGNOSTIC _Pragma("clang diagnostic pop")
 #else
-#define IGNORE_PEDANTIC_WARNINGS _Pragma("GCC diagnostic ignored \"-Wpedantic\"")
-#define END_IGNORE_PEDANTIC_WARNINGS _Pragma("GCC diagnostic pop")
-#define ALLOW_UNUSED_RESULT _Pragma("GCC diagnostic ignored \"-Wunused-result\"")
-#define END_ALLOW_UNUSED_RESULT _Pragma("GCC diagnostic pop")
+#define START_IGNORE_DIAGNOSTIC(diag) _Pragma(xstr(GCC diagnostic push))  _Pragma(xstr(GCC diagnostic ignored diag))
+#define END_IGNORE_DIAGNOSTIC _Pragma("GCC diagnostic pop")
 #endif
+
+#define IGNORE_PEDANTIC_WARNINGS START_IGNORE_DIAGNOSTIC("-Wpedantic")
+#define END_IGNORE_PEDANTIC_WARNINGS END_IGNORE_DIAGNOSTIC
+#define ALLOW_UNUSED_RESULT IGNORE_DIAGNOSTIC("-Wunused-result")
+#define END_ALLOW_UNUSED_RESULT END_IGNORE_DIAGNOSTIC
 #define START_ALLOW_CASE_RANGE IGNORE_PEDANTIC_WARNINGS
 #define END_ALLOW_CASE_RANGE END_IGNORE_PEDANTIC_WARNINGS
 #define BIT_MASK(__TYPE__, __ONE_COUNT__) \
@@ -263,7 +300,7 @@ typedef struct {
 } Cursor;
 
 typedef struct {
-    bool is_visible, is_focused;
+    bool is_visible, is_focused, render_even_when_unfocused;
     CursorShape shape;
     unsigned int x, y;
 } CursorRenderInfo;
@@ -310,9 +347,6 @@ typedef struct {int x;} *SPRITE_MAP_HANDLE;
 #define FONTS_DATA_HEAD SPRITE_MAP_HANDLE sprite_map; double logical_dpi_x, logical_dpi_y, font_sz_in_pts; unsigned int cell_width, cell_height;
 typedef struct {FONTS_DATA_HEAD} *FONTS_DATA_HANDLE;
 
-#define PARSER_BUF_SZ (8 * 1024)
-#define READ_BUF_SZ (1024*1024)
-
 #define clear_sprite_position(cell) (cell).sprite_x = 0; (cell).sprite_y = 0; (cell).sprite_z = 0;
 
 #define ensure_space_for(base, array, type, num, capacity, initial_cap, zero_mem) \
@@ -338,6 +372,8 @@ cursor_to_attrs(const Cursor *c, const uint16_t width) {
     return ans;
 }
 
+#define cursor_as_gpu_cell(cursor) {.attrs=cursor_to_attrs(cursor, 0), .fg=(cursor->fg & COL_MASK), .bg=(cursor->bg & COL_MASK), .decoration_fg=cursor->decoration_fg & COL_MASK}
+
 static inline void
 attrs_to_cursor(const CellAttrs attrs, Cursor *c) {
     c->decoration = attrs.decoration; c->bold = attrs.bold;  c->italic = attrs.italic;
@@ -359,8 +395,8 @@ void cursor_reset(Cursor*);
 Cursor* cursor_copy(Cursor*);
 void cursor_copy_to(Cursor *src, Cursor *dest);
 void cursor_reset_display_attrs(Cursor*);
-void cursor_from_sgr(Cursor *self, int *params, unsigned int count);
-void apply_sgr_to_cells(GPUCell *first_cell, unsigned int cell_count, int *params, unsigned int count);
+void cursor_from_sgr(Cursor *self, int *params, unsigned int count, bool is_group);
+void apply_sgr_to_cells(GPUCell *first_cell, unsigned int cell_count, int *params, unsigned int count, bool is_group);
 const char* cell_as_sgr(const GPUCell *, const GPUCell *);
 const char* cursor_as_sgr(const Cursor *);
 
@@ -369,7 +405,7 @@ bool schedule_write_to_child(unsigned long id, unsigned int num, ...);
 bool schedule_write_to_child_python(unsigned long id, const char *prefix, PyObject* tuple_of_str_or_bytes, const char *suffix);
 bool set_iutf8(int, bool);
 
-DynamicColor colorprofile_to_color(ColorProfile *self, DynamicColor entry, DynamicColor defval);
+DynamicColor colorprofile_to_color(const ColorProfile *self, DynamicColor entry, DynamicColor defval);
 color_type
 colorprofile_to_color_with_fallback(ColorProfile *self, DynamicColor entry, DynamicColor defval, DynamicColor fallback, DynamicColor falback_defval);
 void copy_color_table_to_buffer(ColorProfile *self, color_type *address, int offset, size_t stride);
@@ -391,3 +427,13 @@ SPRITE_MAP_HANDLE alloc_sprite_map(unsigned int, unsigned int);
 SPRITE_MAP_HANDLE free_sprite_map(SPRITE_MAP_HANDLE);
 const char* get_hyperlink_for_id(const HYPERLINK_POOL_HANDLE, hyperlink_id_type id, bool only_url);
 void log_event(const char *format, ...) __attribute__((format(printf, 1, 2)));
+
+#define memset_array(array, val, count) if ((count) > 0) { \
+    (array)[0] = (val); \
+    size_t __copied__ = 1; \
+    while (__copied__ < (count)) { \
+        const size_t __num__ = MIN(__copied__, (count) - __copied__); \
+        memcpy((array) + __copied__, (array), __num__ * sizeof((val))); \
+        __copied__ += __num__; \
+    } \
+}
