@@ -576,10 +576,9 @@ START_ALLOW_CASE_RANGE
         case 0x2500 ... 0x2573:
         case 0x2574 ... 0x259f:
         case 0x2800 ... 0x28ff:
-        case 0xe0b0 ... 0xe0bf:  // powerline box drawing
-        case 0x1fb00 ... 0x1fb8b:  // symbols for legacy computing
-        case 0x1fba0 ... 0x1fbae:
-        case 0x1fb90:  // inverse medium shade
+        case 0xe0b0 ... 0xe0bf:    // powerline box drawing
+        case 0xee00 ... 0xee0b:    // fira code progress bar/spinner
+        case 0x1fb00 ... 0x1fbae:  // symbols for legacy computing
             return BOX_FONT;
         default:
             *is_emoji_presentation = has_emoji_presentation(cpu_cell, gpu_cell);
@@ -614,17 +613,12 @@ START_ALLOW_CASE_RANGE
     switch(ch) {
         case 0x2500 ... 0x259f:
             return ch - 0x2500; // IDs from 0x00 to 0x9f
-        case 0xe0b0 ... 0xe0d4:
-            return 0xa0 + ch - 0xe0b0; // IDs from 0xa0 to 0xc4
-        case 0x1fb00 ... 0x1fb8b:
-            return 0xc5 + ch - 0x1fb00; // IDs from 0xc5 to 0x150
-        case 0x1fba0 ... 0x1fbae:  // IDs from 0x151 to 0x15f
-            return 0x151 + ch - 0x1fba0;
+        case 0xe0b0 ... 0xee0b:
+            return 0xa0 + ch - 0xe0b0;   // IDs from 0xa0 to 0xd58
         case 0x2800 ... 0x28ff:
-            return 0x160 + ch - 0x2800;
-        case 0x1fb90:
-            //  Allocated to allow for 0x1fb8c ... 0x1fb94 eventually
-            return 0x25f + ch - 0x1fb8c;
+            return 0xe00 + ch - 0x2800; // IDs from 0xe00 to 0xeff
+        case 0x1fb00 ... 0x1fbae:
+            return 0xf00 + ch - 0x1fb00; // IDs from 0xf00 to 0xfae
         default:
             return 0xffff;
     }
