@@ -9,7 +9,6 @@
 #include "keys.h"
 #include "screen.h"
 #include "glfw-wrapper.h"
-#include "control-codes.h"
 #include <structmember.h>
 
 // python KeyEvent object {{{
@@ -23,7 +22,7 @@ typedef struct {
 static PyObject* convert_glfw_key_event_to_python(const GLFWkeyevent *ev);
 
 static PyObject*
-new(PyTypeObject *type UNUSED, PyObject *args, PyObject *kw) {
+new_keyevent_object(PyTypeObject *type UNUSED, PyObject *args, PyObject *kw) {
     static char *kwds[] = {"key", "shifted_key", "alternate_key", "mods", "action", "native_key", "ime_state", "text", NULL};
     GLFWkeyevent ev = {.action=GLFW_PRESS};
     if (!PyArg_ParseTupleAndKeywords(args, kw, "I|IIiiiiz", kwds, &ev.key, &ev.shifted_key, &ev.alternate_key, &ev.mods, &ev.action, &ev.native_key, &ev.ime_state, &ev.text)) return NULL;
@@ -70,7 +69,7 @@ PyTypeObject PyKeyEvent_Type = {
     .tp_flags = Py_TPFLAGS_DEFAULT,
     .tp_doc = "A key event",
     .tp_members = members,
-    .tp_new = new,
+    .tp_new = new_keyevent_object,
 };
 
 static PyObject*
@@ -102,13 +101,25 @@ update_ime_focus(OSWindow *osw, bool focused) {
 }
 
 void
-update_ime_position(Window* w, Screen *screen) {
-    unsigned int cell_width = global_state.callback_os_window->fonts_data->cell_width, cell_height = global_state.callback_os_window->fonts_data->cell_height;
+prepare_ime_position_update_event(OSWindow *osw, Window *w, Screen *screen, GLFWIMEUpdateEvent *ev) {
+    unsigned int cell_width = osw->fonts_data->cell_width, cell_height = osw->fonts_data->cell_height;
     unsigned int left = w->geometry.left, top = w->geometry.top;
-    left += screen->cursor->x * cell_width;
-    top += screen->cursor->y * cell_height;
+    if (screen_is_overlay_active(screen)) {
+        left += screen->overlay_line.cursor_x * cell_width;
+        top += MIN(screen->overlay_line.ynum + screen->scrolled_by, screen->lines - 1) * cell_height;
+    } else {
+        left += screen->cursor->x * cell_width;
+        top += screen->cursor->y * cell_height;
+    }
+    ev->cursor.left = left; ev->cursor.top = top; ev->cursor.width = cell_width; ev->cursor.height = cell_height;
+}
+
+void
+update_ime_position(Window* w UNUSED, Screen *screen UNUSED) {
     GLFWIMEUpdateEvent ev = { .type = GLFW_IME_UPDATE_CURSOR_POSITION };
-    ev.cursor.left = left; ev.cursor.top = top; ev.cursor.width = cell_width; ev.cursor.height = cell_height;
+#ifndef __APPLE__
+    prepare_ime_position_update_event(global_state.callback_os_window, w, screen, &ev);
+#endif
     glfwUpdateIMEState(global_state.callback_os_window->handle, &ev);
 }
 
@@ -154,12 +165,12 @@ on_key_input(GLFWkeyevent *ev) {
         case GLFW_IME_WAYLAND_DONE_EVENT:
             // If we update IME position here it sends GNOME's text input system into
             // an infinite loop. See https://github.com/kovidgoyal/kitty/issues/5105
-            screen_draw_overlay_text(screen, NULL);
+            screen_update_overlay_text(screen, NULL);
             debug("handled wayland IME done event\n");
             return;
         case GLFW_IME_PREEDIT_CHANGED:
+            screen_update_overlay_text(screen, text);
             update_ime_position(w, screen);
-            screen_draw_overlay_text(screen, text);
             debug("updated pre-edit text: '%s'\n", text);
             return;
         case GLFW_IME_COMMIT_TEXT:
@@ -167,7 +178,7 @@ on_key_input(GLFWkeyevent *ev) {
                 schedule_write_to_child(w->id, 1, text, strlen(text));
                 debug("committed pre-edit text: %s\n", text);
             } else debug("committed pre-edit text: (null)\n");
-            screen_draw_overlay_text(screen, NULL);
+            screen_update_overlay_text(screen, NULL);
             return;
         case GLFW_IME_NONE:
             // for macOS, update ime position on every key input
@@ -188,17 +199,6 @@ on_key_input(GLFWkeyevent *ev) {
     else { consumed = ret == Py_True; Py_CLEAR(ret); } \
     w = window_for_window_id(active_window_id); \
 }
-    if (global_state.in_sequence_mode) {
-        debug("in sequence mode, handling as a potential shortcut\n");
-        dispatch_key_event(process_sequence);
-        if (dispatch_ok) {
-          if (consumed && action != GLFW_RELEASE && w && !is_modifier_key(key)) {
-            w->last_special_key_pressed = key;
-          }
-        }
-        return;
-    }
-
     if (action == GLFW_PRESS || action == GLFW_REPEAT) {
         w->last_special_key_pressed = 0;
         dispatch_key_event(dispatch_possible_special_key);
