@@ -219,11 +219,11 @@ def set_cocoa_global_shortcuts(opts: Options) -> Dict[str, SingleKey]:
     if is_macos:
         from collections import defaultdict
         func_map = defaultdict(list)
-        for k, v in opts.keyboard_modes[''].keymap.items():
-            for kd in v:
-                if kd.is_suitable_for_global_shortcut:
-                    parts = tuple(kd.definition.split())
-                    func_map[parts].append(k)
+        for single_key, v in opts.keyboard_modes[''].keymap.items():
+            kd = v[-1]  # the last definition is the active one
+            if kd.is_suitable_for_global_shortcut:
+                parts = tuple(kd.definition.split())
+                func_map[parts].append(single_key)
 
         for ac in ('new_os_window', 'close_os_window', 'close_tab', 'edit_config_file', 'previous_tab',
                    'next_tab', 'new_tab', 'new_window', 'close_window', 'toggle_macos_secure_keyboard_entry', 'toggle_fullscreen',
@@ -363,7 +363,7 @@ def macos_cmdline(argv_args: List[str]) -> List[str]:
     ans = list(shlex_split(raw))
     if ans and ans[0] == 'kitty':
         del ans[0]
-    return ans
+    return ans + argv_args
 
 
 def expand_listen_on(listen_on: str, from_config_file: bool) -> str:
